#!/usr/bin/env python

############
#
#    This program downloads a specific list of datasets from a given history.
#    The syntax is:
#       python download_datasets_cmd.py [History Name or ID] ['name' or 'ID'] [Local Folder Name] [List of Datasets to DL]
#
#    WARNING! To use this script, you must update the server URL in line 61 and your API Id in line 62
#

### Import basic libraries
import json
import sys
import os
import datetime

### Import BioBlend library
from bioblend import galaxy
from bioblend.galaxy.libraries import *
from bioblend.galaxy.client import *
from bioblend.galaxy.objects import *
from bioblend.galaxy.workflows import *
from bioblend.galaxy.histories import *
from bioblend.galaxy.datasets import *


def main():

    input_settings = configure_settings()

    gi = galaxy.GalaxyInstance(url=input_settings['galaxy_url'], key=input_settings['api_key'])

    ### Create output directory
    if not os.path.exists(input_settings['output_dir']):
        os.makedirs(input_settings['output_dir'])

    ### Get JSON file of history information
    history_info = get_history_info(gi,input_settings['history_name'],input_settings['history_id'])

    ### Set history and id variables
    input_settings['history_id'] = history_info[0]['id']
    input_settings['history_name'] = history_info[0]['name']

    ### Get list of history datasets
    history_datasets = get_history_datasets(gi,input_settings['history_id'])

    ### Create list of dataset names to download
    datasets_to_dl = create_downloads_list(input_settings['dataset_list'])

    ### Save datasets to disk
    save_datasets_to_file(gi,history_datasets,datasets_to_dl,input_settings['output_dir'])

    print input_settings['output_dir']

    sys.exit(0)

def configure_settings():
    ### Parse the command line arguments for the settings
    input_settings = {
        'galaxy_url':'http://128.8.185.39/',
        'api_key':'7db6cdc69134bfe23a31d7501886f3e6',
        'dataset_list':sys.argv[4]
        }

    input_settings['output_dir'] = sys.argv[3] #+'_'+str(datetime.datetime.today().strftime("%d-%m-%Y_%H:%M:%S"))+'/'
    #import_input_file = False

    ### Set history name or id
    if sys.argv[2] is 'name':
        input_settings['history_name'] = sys.argv[1]
        input_settings['history_id'] = None
    else:
        input_settings['history_name'] = None
        input_settings['history_id'] = sys.argv[1]

    return input_settings


def save_datasets_to_file(gi,history_datasets,datasets_to_dl,output_dir):
    ### Save the datasets to a local file
    for i in history_datasets:
        if i['name'] in datasets_to_dl:
            gi.datasets.download_dataset(i['id'],file_path=(output_dir+'/'+i['name'].split('.')[0]), use_default_filename=False)


def get_history_info(gi,history_name,history_ID):
    ### Get the history info using the name or ID
    if history_ID is None:
        return gi.histories.get_histories(name=history_name)
    else:
        return gi.histories.get_histories(history_id=history_ID)


def get_history_datasets(gi,history_id):
    ### Get a list of datasets in the history
    dataset_list = gi.histories.show_matching_datasets(history_id)
    #print "list of datasets: ", dataset_list
    new_list = []
    for i in dataset_list:
        if i['state'] == 'ok':
            new_list.append(i)
    #print "new_list: ", new_list
    return new_list


def create_downloads_list(dataset_list):
    ### Create a list of datasets to download
    datasets_to_dl = []
    data = open(dataset_list)
    for line in data:
        datasets_to_dl.append(line.strip('\n'))
    return datasets_to_dl


if __name__ == "__main__":
    main()
