#!/usr/bin/env python

############
#
#    This program runs a series of workflows on a specified dataset
#    The syntax is:
#       python run_workflows8.py [label for this run] [name of JSON file] [input dataset1 ID] [input dataset1 label] [input dataset2 ID] [input dataset2 label] ...
#
# python run_workflows_new.py SRX10101 input_settings.json bbd44e69cb8906b5e6fb41fe95b30581 FASTQ bbd44e69cb8906b54eeca3a85d035a85 miRNA bbd44e69cb8906b5640408f4aff7edea Ce6
#
###

### Import basic libraries
import json
import yaml
import sys
import os
import csv
import datetime
import time as time
import subprocess

### Import BioBlend library
from bioblend import galaxy
from bioblend.galaxy.libraries import *
from bioblend.galaxy.client import *
from bioblend.galaxy.objects import *
from bioblend.galaxy.workflows import *
from bioblend.galaxy.histories import *
from bioblend.galaxy.datasets import *
from bioblend.galaxy.tools import *

### Print command line arguments
print sys.argv

### Determine the time this run was started
start_time = str(datetime.datetime.today().strftime("%d-%m-%Y_%H:%M:%S"))

def main():

    settings_file = load_settings()
    settings_file['data_label'] = sys.argv[1]
    gi = galaxy_setup(settings_file)
    gi.verify = False
    settings_file = cmd_settings(settings_file)

    dl_folder_name = settings_file['data_label'] + settings_file['gi_setup']['output_dir']+'/'

    ### Create output directory
    if not os.path.exists(dl_folder_name):
        os.makedirs(dl_folder_name)
 
    # Print the static data
    print "settings_file static data: ", settings_file['static_data']

    wf_results = run_workflows(gi,settings_file)

    print wf_results
    print 'stuff to download', wf_results[1]
    #print " list: ", gi.histories.show_history(wf_results[0]['CCC']['history_id'],contents=True)

    download_datasets(gi,wf_results[1],settings_file['gi_setup']['wait_for_history'],dl_folder_name)
	
    print "Workflow complete!"
    sys.exit(0)

def galaxy_setup(settings_file):
    ### Create a new Galaxy Instance for BioBlend
    return galaxy.GalaxyInstance(url=settings_file['gi_setup']['galaxy_url'], key=settings_file['gi_setup']['api_key'])


def cmd_settings(settings_file):
    ### Parse the command line arguments
    for i in range(0,len(sys.argv)-3,2):
        print i
        settings_file['static_data'][sys.argv[i+4]] = sys.argv[i+3]

    print settings_file['static_data']
    return settings_file


def get_history_dataset_list(gi,history_id):
    ### Get the list of datasets in a history
    return gi.histories.show_history(history_id,contents=True)


def run_workflows(gi,settings_file):
    ### Run a new workflow
    wf_results = []
    wf_outputs = []
    print "settings file: ", settings_file
    for i in range(len(settings_file['workflow_settings'].keys())):
        print 'i: ',i
        print 'i contents: ', settings_file['workflow_settings'][str(i)]
        for wf in settings_file['workflow_settings'][str(i)]:
            workflow_inputs = get_workflow_inputs(gi,wf['id'])
            print 'workflow_inputs: ', workflow_inputs
            param_set = create_tool_parameter_set(wf)
            print 'param_set: ', param_set
            dataset_map = create_dataset_map(gi,wf,workflow_inputs,settings_file['static_data'],wf_results)
            print 'dataset map: ', dataset_map[0]
            print 'histories to wait for: ', dataset_map[1]
            for history_id in dataset_map[1]:
                wait_for_history(gi,history_id,settings_file['gi_setup']['wait_for_history'])
            
	    print "settings_file['data_label']: ", settings_file['data_label']
	    print "wf['tag']: ", wf['tag']
            workflow_name = get_workflow_name(gi, wf['id'])
            print "get_workflow_name(gi, wf['id']): ", workflow_name
            history_name=(settings_file['data_label']+'_'+wf['tag']+'_'+workflow_name)
            print "wf['id']: ", wf['id']
            print "dataset_map[0]: ", dataset_map[0]
            print "param_set: ", param_set
            print "history_name: ", history_name
            print "settings_file['gi_setup']['import_input_file']: ", settings_file['gi_setup']['import_input_file']
            try:
                wf_return = gi.workflows.run_workflow(wf['id'],
                    dataset_map=dataset_map[0],
                    params=param_set,
                    history_name=history_name,
                    import_inputs_to_history=settings_file['gi_setup']['import_input_file']
                    )
            except ConnectionError, e:
                print "caught an exception: connection error: ", e
                continue
            # Wait for this history to finish before continuing
            #print "wf_return: ", wf_return
            #wait_for_history(gi,wf_return['history'],settings_file['gi_setup']['wait_for_history'])
            
            #print "wf_return: ", wf_return, "wf_return history_id: ", wf_return['history_id'],"wf_return id: ", wf_return['id']
            print "wf.keys(): ", wf.keys()
            if 'tag' in wf.keys():
                wf_results.append({wf['tag']:get_history_dataset_list(gi,wf_return['history'])})
                print "wf_results: ", wf_results
            write_to_output(settings_file, [settings_file['data_label'], wf_return['history'], history_name])
            if 'wf_outputs' in wf.keys():
                print 'wf_outputs: ', wf['wf_outputs']
                print 'wf_return: ', wf_return
                for output in wf['wf_outputs']:
                    print 'output: ', output
                    wf_outputs.append({'history_id':wf_return['history'],'history_name':history_name,'dl_datasets':output,'tag':wf['tag']})
            last_history = wf_return['history']
        wait_for_history(gi,last_history,settings_file['gi_setup']['wait_for_history'])         
    return (wf_results, wf_outputs)

def get_workflow_name(gi, wf_id):
    ### Get a workflow name using the ID
    return gi.workflows.show_workflow(wf_id)['name']


def get_workflow_inputs(gi,wf_id):
    ### Get a workflows inputs using the ID
    return gi.workflows.show_workflow(wf_id)['inputs']


def get_dataset_name(gi,dataset_id):
    ### Get a dataset name using the ID
    return gi.datasets.show_dataset(dataset_id)['name']


def create_tool_parameter_set(wf_settings):
    ### Create the BioBlend syntax to change a parameter when running a tool
    if 'params' not in wf_settings.keys():
        return None

    param_set = dict()
    for i in wf_settings['params']:
        step_id = i['step_id']
        #tool_id = i['tool_id']
        param_input = i['param_input']
        #param_set[tool_id] = param_input
        param_set[step_id] = param_input
    return param_set


def create_dataset_map(gi,wf_settings,workflow_inputs,static_datasets,wf_results):
    ### Create a dict() map of all the datasets required to start a workflow
    dataset_map = dict()
    history_ids_to_wait_for = []
    for input_properties in wf_settings['wf_inputs']:
        print "input properties: ", input_properties['source']
        if input_properties['source'] == "static_data":
            dataset_id = static_datasets[input_properties['data_name']]
            print "input_id: ", dataset_id
        if input_properties['source'] == "history":
            input_tag = input_properties['source_tag']
            print 'input_tag: ', input_tag
            print 'wf_results: ', wf_results
            for result in wf_results:
                print "ds.keys(): ", result.keys()
                if input_tag in result.keys():
                    print 'result: ', result
                    for dataset in result[input_tag]:
                        if dataset['name'] == input_properties['data_name']:
                            dataset_id = dataset['id']
                            dataset_name = dataset['name']
                            history_ids_to_wait_for.append(dataset['history_id'])
        for i in workflow_inputs.keys():
            print i
            print 'workflow_inputs[i]["label"]', workflow_inputs[i]['label']
            if workflow_inputs[i]['label'] == input_properties['wf_step_name']:
                workflow_input_id = i
        dataset_src = get_dataset_source(gi,dataset_id)
        dataset_map[workflow_input_id] = {'id':dataset_id,'src':dataset_src}

    print 'dataset_map - new: ',dataset_map
    #print history_ids_to_wait_for
    #print set(history_ids_to_wait_for)
    return (dataset_map,set(history_ids_to_wait_for))


def wait_for_history(gi,history_id,delay):
    ### Waits a specified amount of time for a dependent workflow to complete before using one of its datasets as an input
    history_status = gi.histories.get_status(history_id)
    while (((history_status['state_details']['paused'] + 
			history_status['state_details']['upload'] + 
			history_status['state_details']['running'] + 
			history_status['state_details']['setting_metadata'] + 
			history_status['state_details']['new'] + 
			history_status['state_details']['queued'] + 
			history_status['state_details']['failed_metadata']) is not 0) or (int(history_status['state_details']['ok']) <= 10)):
        print 'history_status : ', history_status, str(datetime.datetime.today().strftime("%d-%m-%Y_%H:%M:%S"))
        time.sleep(int(delay))
        history_status = gi.histories.get_status(history_id)
    print 'history_status - percent complete: ', history_status


def get_dataset_source(gi,dataset_id):
    ### Identify the type of dataset to be used as input. BioBlend requires this to use a dataset as input
    print gi.datasets.show_dataset(dataset_id)
    return gi.datasets.show_dataset(dataset_id)['hda_ldda']


def load_settings():
    ### Loads a JSON file that describes the overall workflow
    data = open(sys.argv[2])
	#settings_JSON = json.load(data)
    settings_JSON = yaml.load(data)
    if settings_JSON['gi_setup']['import_input_file'] == "True":
        settings_JSON['gi_setup']['import_input_file'] = True
    if settings_JSON['gi_setup']['import_input_file'] == "False":
        settings_JSON['gi_setup']['import_input_file'] = False
    print "json settings: ", settings_JSON
    return settings_JSON
	

def download_datasets(gi,dl_datasets,delay,dl_folder_name):
    ### Downloads specified datasets from the Galaxy server to a local directory
    for history in dl_datasets:
        print "history", history
        history_id = history['history_id']
        wait_for_history(gi,history_id,delay)
        folder_name = dl_folder_name+history['tag']+'/'
        if history['dl_datasets']['output_source'] == 'file':
            dl_filename = history['dl_datasets']['filename']
        try:
            subprocess_params = ['python','download_datasets_cmd.py',str(history_id), 'ID',str(folder_name), str(dl_filename)]
        except IOError, e:
            print "IOError: ", e
        print 'subprocess_params: ',subprocess_params
        dl_output = run_subprocess(subprocess_params,[False])
        print 'dl output: ',  dl_output
        print "history keys: ", history.keys()
        if 'run_cmd' in history['dl_datasets'].keys():
            for cmd in history['dl_datasets']['run_cmd']:
                run_cmd(gi,folder_name,cmd)


def run_cmd(gi,folder_name,run_cmd_list):
    ### Starts the process of running a command line program on a downloaded dataset
    subprocess_cwd = [False]
    subprocess_params = []
    print 'param list: ', run_cmd_list['param_list']
    for param in run_cmd_list['param_list']:
        print 'param: ', param
        if param == 'FOLDER_NAME':
            print 'folder name: ',folder_name
            subprocess_params.append(folder_name)
        elif param == 'FILE_NAME':
            print 'FILE_NAME:'          
            file_name_index = len(subprocess_params)
            subprocess_params.append('FILE_NAME')
        elif param == 'CWD':
            subprocess_cwd[0] = True
            subprocess_cwd.append(folder_name)
        else:
            #param != 'FOLDER_NAME' and param != 'FILE_NAME':
            subprocess_params.append(run_cmd_list[param])

    print 'initial arg_list: ', subprocess_params

    if 'input_file_source' in run_cmd_list:
        if run_cmd_list['input_file_source'] == 'dir_list':
            file_list = list_directory(folder_name)
            print "dir_list: ", file_list
        elif run_cmd_list['input_file_source'] == 'txt_list':
            file_list = parse_data_to_list(run_cmd_list['input_file_list'])
            print "txt_list", file_list
        else:
            file_list = run_cmd_list['input_file_list']
            print "input file list: ", file_list

        for file_name in file_list:
            print "file name to run: ", file_name
            subprocess_params[file_name_index] = file_name
            print "subprocess params: ", subprocess_params        
            try:
                run_cmd_output = run_subprocess(subprocess_params,subprocess_cwd)
                print run_cmd_output 
            except Exception, e:
                print "Error running output: ", e     
    else:
        try:
            run_cmd_output = run_subprocess(subprocess_params,subprocess_cwd)
        except Exception, e:
            print "Error running output: ", e

def parse_data_to_list(file_name):
    ### This function is used to convert a text list of elements to a python list. The input is the opened data, and the output is a list of items.
    new_list = []
    data = open(file_name)
    for line in data:
        new_list.append(line.strip('\n'))
    return new_list


def list_directory(dir_name):
    ### Determine the contents of a directory
    file_list_text = os.listdir(dir_name)
    file_list = []
    for line in file_list_text:
        file_list.append(line.strip('\n'))
    print 'This is the directory contents: ', file_list
    return file_list

def run_subprocess(arguments,cwd_dir):
    ### Run a subprocess on a downloaded data file
    print 'running subprocess with arguments: ', arguments
    try:
		if cwd_dir[0] is True:
		    print "cwd_dir[1]", cwd_dir[1]
		    subprocess_output = subprocess.check_output(arguments,cwd=cwd_dir[1])
		else:
		    subprocess_output = subprocess.check_output(arguments)	
    except Exception, e:
        print "Error running command on file: ", e
    
    return subprocess_output

def write_to_output(settings_file, wf_output):
    ### Outputs the histories created in this run to a text file. 
    
    # Create writer
    ofile  = open(settings_file['data_label'] + settings_file['gi_setup']['output_dir']+'/'+ wf_output[0] + "_" + start_time + '.log', "a")
    writer = csv.writer(ofile, delimiter='\t')

    line = wf_output[1:]
    writer.writerow(line)
    writer.writerow('')

    # Close the file
    ofile.close()
    return



if __name__ == "__main__":
    main()
